import sys
import warnings
from weaviate.exceptions import RequestsConnectionError, UnexpectedStatusCodeException
from weaviate.connect import REST_METHOD_POST, Connection
from .requests import BatchRequest, ObjectsBatchRequest, ReferenceBatchRequest

class Batch:
    """
    Batch class used to add multiple objects or object references at once into weaviate.
    """

    def __init__(self,connection: Connection):
        """
        Initialize a Batch class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    def create(self, batch_request: BatchRequest) -> list:
        """
        Load data objects in batches, either Objects or References.
        Loading batch References is faster by ignoring some validations.
        Loading inconsistent data may ends up in an invalid graph.
        If the consistency of the References is wanted use
        'Client().data_object.reference.add' to have additional validation instead.

        Parameters
        ----------
        batch_request : weaviate.batch.BatchRequest
            Contains all the data objects that should be added in one batch.
            Note: Should be a sub-class of BatchRequest since BatchRequest
            is just an abstract class.

        Returns
        -------
        list
            A list with the status of every data object added.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if isinstance(batch_request, ObjectsBatchRequest):
            data_object_type = "objects"
        elif isinstance(batch_request, ReferenceBatchRequest):
            data_object_type = "references"
        else:
            raise TypeError("Wrong argument type, expected a sub-class of BatchRequest \
                    (ObjectsBatchRequest or ReferenceBatchRequest), got: " +\
                    str(type(batch_request)))

        path = f"/batch/{data_object_type}"

        try:
            response = self._connection.run_rest(
                path=path,
                rest_method=REST_METHOD_POST,
                weaviate_object=batch_request.get_request_body()
                )
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                        + ' Connection error, batch was not added to weaviate.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 200:
            return response.json()
        raise UnexpectedStatusCodeException(f"Create {data_object_type} in batch", response)

    def create_objects(self, objects_batch_request: ObjectsBatchRequest) -> list:
        """
        Creates multiple objects at once in weaviate

        Parameters
        ----------
        objects_batch_request : weaviate.batch.ObjectsBatchRequest
            The batch of objects that should be added.

        Returns
        -------
        list
            A list with the status of every thing that was created.
        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if not isinstance(objects_batch_request, ObjectsBatchRequest):
            raise TypeError(f"'reference_batch_request' should be of type \
                ObjectsBatchRequest but was given : {type(objects_batch_request)}")

        return self.create(
            batch_request=objects_batch_request
            )

    def create_references(self, reference_batch_request: ReferenceBatchRequest) -> list:
        """
        Batch loading references.
        Loading batch references is faster by ignoring some validations.
        Loading inconsistent data may ends up in an invalid graph.
        If the consistency of the references is wanted use
        'Client().data_object.reference.add' to have additional validation instead.

        Parameters
        ----------
        reference_batch_request : weaviate.batch.ReferenceBatchRequest
            Contains all the references that should be added in one batch.

        Returns
        -------
        list
            A list with the status of every reference added.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if not isinstance(reference_batch_request, ReferenceBatchRequest):
            raise TypeError(f"'reference_batch_request' should be of type \
                ReferenceBatchRequest but was given : {type(reference_batch_request)}")

        return self.create(
            batch_request=reference_batch_request
            )

    def add_references(self, reference_batch_request: ReferenceBatchRequest) -> list:
        """
        'add_references' is deprecated, use 'create' or 'create_references' instead!
        """

        warnings.warn(
            "'add_references' is deprecated, use 'create' or 'create_references' instead!",
            DeprecationWarning
        )
        return self.create_references(
            reference_batch_request=reference_batch_request
            )



import sys
from typing import Union
from weaviate.connect import Connection, REST_METHOD_POST, REST_METHOD_GET, REST_METHOD_DELETE
from weaviate.util import _get_dict_from_object, _is_sub_schema
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from weaviate.schema.validate_schema import validate_schema, check_class
from weaviate.schema.properties import Property


_PRIMITIVE_WEAVIATE_TYPES_SET = set(["string", "int", "boolean", "number", "date", "text",\
                                                            "geoCoordinates", "CrossRef"])


class Schema:
    """
    Schema class used to interact and manipulate schemas or classes.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Schema class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection
        self.property = Property(self._connection)

    def create(self, schema: Union[dict, str]) -> None:
        """
        Create the schema at the weaviate instance.

        Parameters
        ----------
        schema : dict or str
            Schema as a python dict, or the path to a json file or a url of a json file.

        Raises
            TypeError
                If the 'schema' is neither a string nor a dict.
            ValueError
                If 'schema' can not be converted into a weaviate schema.
            ConnectionError
                If the network connection to weaviate fails.
            UnexpectedStatusCodeException
                If weaviate reports a none OK status.
            SchemaValidationException
                If the 'schema' could not be validated against the standard format.
        """

        loaded_schema = _get_dict_from_object(schema)
        # validate the schema before loading
        validate_schema(loaded_schema)
        self._create_classes_with_primitives(loaded_schema["classes"])
        self._create_complex_properties_from_classes(loaded_schema["classes"])

    def create_class(self, schema_class: Union[dict, str]) -> None:
        """
        Create a single class as part of the schema in weaviate.

        Parameters
        ----------
        schema_class : dict or str
            Class as a python dict, or the path to a json file or a url of a json file.

        Raises
            TypeError
                If the 'schema_class' is neither a string nor a dict.
            ValueError
                If 'schema_class' can not be converted into a weaviate schema.
            ConnectionError
                If the network connection to weaviate fails.
            UnexpectedStatusCodeException
                If weaviate reports a none OK status.
            SchemaValidationException
                If the 'schema_class' could not be validated against the standard format.
        """

        loaded_schema_class = _get_dict_from_object(schema_class)
        # validate the class before loading
        check_class(loaded_schema_class)
        self._create_class_with_premitives(loaded_schema_class)
        self._create_complex_properties_from_class(loaded_schema_class)

    def delete_class(self, class_name: str) -> None:
        """
        Delete a schema class from weaviate. This deletes all associated data.

        Parameters
        ----------
        class_name : str
            The class that should be deleted from weaviate.

        Raises
        ------
        TypeError
            If 'class_name' argument not of type str.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if not isinstance(class_name, str):
            raise TypeError(f"Class name was {type(class_name)} instead of str")

        path = f"/schema/{class_name}"
        try:
            response = self._connection.run_rest(path, REST_METHOD_DELETE)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, during deletion of class.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Delete class from schema", response)

    def delete_all(self) -> None:
        """
        Remove the entire schema from the weavaite instance and all data associated with it.
        """

        schema = self.get()
        classes = schema.get("classes", [])
        for _class in classes:
            self.delete_class(_class["class"])


    def contains(self, schema: dict=None) -> bool:
        """
        Check if weaviate already contains a schema.

        Parameters
        ----------
        schema : dict, optional
            If a schema is given it is checked if this specific schema is already loaded.
            It will test only this schema. If the given schema is a subset of the loaded
            schema it will still return true, by default None.

        Returns
        -------
        bool
            True if a schema is present,
            False otherwise.
        """

        loaded_schema = self.get()

        if schema is not None:
            sub_schema = _get_dict_from_object(schema)
            return _is_sub_schema(sub_schema, loaded_schema)

        if len(loaded_schema["classes"]) == 0:
            return False
        return True

    def get(self) -> dict:
        """
        Get the schema from weaviate.

        Returns
        -------
        dict
            A dict containing the schema. The schema may be empty.
            To see if a schema has already been loaded use `contains` method.
        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        try:
            response = self._connection.run_rest("/schema", REST_METHOD_GET)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, schema could not be retrieved.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Get schema", response)
        return response.json()

    def _create_complex_properties_from_class(self, schema_class: dict) -> None:
        """
        Add crossreferences to already existing class.

        Parameters
        ----------
        schema_class : dict
            Description of the class that should be added.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if "properties" not in schema_class:
            # Class has no properties nothing to do
            return
        for property_ in schema_class["properties"]:

            if _property_is_primitive(property_["dataType"]):
                continue
            # TODO adapt the schema properties according to the new standard
            # create the property object
            schema_property = {
                "dataType": property_["dataType"],
                "description": property_["description"],
                "name": property_["name"]
            }

            if "index" in property_:
                schema_property["index"] = property_["index"]
            if "vectorizePropertyName" in property_:
                schema_property["vectorizePropertyName"] = property_["vectorizePropertyName"]

            path = "/schema/" + schema_class["class"] + "/properties"
            try:
                response = self._connection.run_rest(path, REST_METHOD_POST, schema_property)
            except RequestsConnectionError as conn_err:
                message = str(conn_err)\
                        + ' Connection error, property may not have been created properly.'
                raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
            if response.status_code != 200:
                raise UnexpectedStatusCodeException("Add properties to classes", response)

    def _create_complex_properties_from_classes(self, schema_classes_list: list) -> None:
        """
        Add crossreferences to already existing classes.

        Parameters
        ----------
        schema_classes_list : list
            A list of classes as they are found in a schema json description.
        """

        for schema_class in schema_classes_list:
            self._create_complex_properties_from_class(schema_class)

    def _create_class_with_premitives(self, weaviate_class: dict) -> None:
        """
        Create class with only primitives.

        Parameters
        ----------
        weaviate_class : dict
            A single weaviate formated class

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        # TODO addapt to the new class schema config
        # Create the class
        schema_class = {
            "class": weaviate_class['class'],
            "properties": []
        }

        if "description" in weaviate_class:
            schema_class['description'] = weaviate_class['description']

        if "vectorizeClassName" in weaviate_class:
            schema_class["vectorizeClassName"] = weaviate_class["vectorizeClassName"]

        if "properties" in weaviate_class:
            schema_class["properties"] = _get_primitive_properties(
                                                    weaviate_class["properties"])

        # Add the item
        try:
            response = self._connection.run_rest("/schema", REST_METHOD_POST, schema_class)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, class may not have been created properly.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Create class", response)

    def _create_classes_with_primitives(self, schema_classes_list: list) -> None:
        """
        Create all the classes in the list and primitive properties.
        This function does not create references,
        to avoid references to classes that do not yet exist.

        Parameters
        ----------
        schema_classes_list : list
            A list of classes as they are found in a schema json description.
        """

        for weaviate_class in schema_classes_list:
            self._create_class_with_premitives(weaviate_class)


def _property_is_primitive(data_type_list: list) -> bool:
    """
    Check if the property is primitive.

    Parameters
    ----------
    data_type_list : list
        Data types to be checkedif are primitive.

    Returns
    -------
    bool
        True if it only consists of primitive data types,
        False otherwise.
    """

    if len(set(data_type_list) - _PRIMITIVE_WEAVIATE_TYPES_SET) == 0:
        return True
    return False


def _get_primitive_properties(properties_list: list) -> list:
    """
    Filter the list of properties for only primitive properties.

    Parameters
    ----------
    properties_list : list
        A list of properties to exctract the primitive properties.

    Returns
    -------
    list
        A list of properties containing only primitives.
    """

    primitive_properties = []
    for property_ in properties_list:
        if not _property_is_primitive(property_["dataType"]):
            # property is complex and therefore will be ignored
            continue
        # TODO check it is fine to add the whole dict
        primitive_properties.append(property_)
    return primitive_properties



import sys
from weaviate.exceptions import RequestsConnectionError, UnexpectedStatusCodeException
from weaviate.connect import REST_METHOD_POST, REST_METHOD_GET, Connection


class Contextionary:
    """
    Contextionary class used to add exted the weaviate contextionary module
    or to get vector/s of a specific concept.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Contextionary class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    def extend(self,
            concept: str,
            definition: str,
            weight: float=1.0
        ) -> None:
        """
        Extend the text2vec-contextionary with new concepts

        Parameters
        ----------
        concept : str
            The new concept that should be added that is not in the weaviate
            or needs to be updated, e.g. an abbreviation.
        definition : str
            The definition of the new concept.
        weight : float, optional
            The weight of the new definition compared to the old one,
            must be inbetween the interval [0.0; 1.0], by default 1.0

        Raises
        ------
        TypeError
            If an argument is not of an approptiate type.
        ValueError
            If 'weight' is outside the interval [0.0; 1.0].
        requests.exceptions.ConnectionError
            If text2vec-contextionary could not be extended.
        weaviate.UnexpectedStatusCodeException
            If the network connection to weaviate fails.
        """

        if not isinstance(concept, str):
            raise TypeError("Concept must be string")
        if not isinstance(definition, str):
            raise TypeError("Definition must be string")
        if not isinstance(weight, float):
            raise TypeError("Weight must be float")

        if weight > 1.0 or weight < 0.0:
            raise ValueError("Weight out of limits 0.0 <= weight <= 1.0")

        extension = {
            "concept": concept,
            "definition": definition,
            "weight": weight
        }

        try:
            response = self._connection.run_rest(
                "/modules/text2vec-contextionary/extensions/",
                REST_METHOD_POST,
                extension
                )
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                    + ' Connection error, text2vec-contextionary could not be extended.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 200:
            # Successfully extended
            return
        raise UnexpectedStatusCodeException("Extend text2vec-contextionary", response)

    def get_concept_vector(self, concept: str) -> dict:
        """
        Retrieves the vector representation of the given concept.

        Parameters
        ----------
        concept : str
            Concept for which the vector should be retrieved.
            May be CamelCase for word combinations.

        Returns
        -------
        dict
            A dictionary containing info and the vector/s of the concept.
            The vector might be empty if the text2vec-contextionary does not contain it.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        Exception
            Unexpected exception that should be reported in an issue.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        AttributeError
        """

        path = "/modules/text2vec-contextionary/concepts/" + concept
        try:
            response = self._connection.run_rest(path, REST_METHOD_GET)
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                    + ' Connection error, text2vec-contextionary vector was not retrieved.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        except AttributeError:
            raise
        except Exception as error:
            message = str(error)\
                    + ' Unexpected exception please report this excetpion in an issue.'
            raise type(error)(message).with_traceback(sys.exc_info()[2])
        else:
            if response.status_code == 200:
                return response.json()
            raise UnexpectedStatusCodeException("text2vec-contextionary vector", response)



import requests
from unittest.mock import Mock



def add_run_rest_to_mock(mock, return_json=None, status_code=200):
    """Adds the run_rest method to a mock using the given options

    :param mock: the mock to add the method to (this is call by reference so this object will be edited)
    :param return_json: the json it should return
    :param status_code: the code it should return
    :return: returns the same mock as given. The object is not copied so the usage of the return value is not necessary
    """
    # Create mock
    return_value_mock = Mock()
    # set the json
    return_value_mock.json.return_value = return_json
    # Set status code
    return_value_mock.configure_mock(status_code=status_code)

    # Add the return value to the given mock
    mock.run_rest.return_value = return_value_mock  # set return object to mock object

    return mock


def run_rest_raise_connection_error(path, rest_method, weaviate_object=None, retries=3, params={}):
    """ A mock that no mather the input will throw an ConnectionError

    :raises ConnectionError
    """
    raise requests.exceptions.ConnectionError


def replace_connection(weaviate, connection):
    """

    :param weaviate:
    :type weaviate: weaviate.Client
    :param connection:
    :type connection: weaviate.connection.Connection
    :return:
    """

    weaviate._connection = connection
    weaviate.classification._connection = connection
    weaviate.schema._connection = connection
    weaviate.schema.property._connection = connection
    weaviate.contextionary._connection = connection
    weaviate.batch._connection = connection
    weaviate.data_object._connection = connection
    weaviate.data_object.reference._connection = connection
    weaviate.query._connection = connection


"""
Weaviate Exceptions.
"""
# Import requests ConnectionError as weaviate.ConnectionError to overwrite buildins connection error
from requests.exceptions import ConnectionError as RequestsConnectionError
from requests import Response


class UnexpectedStatusCodeException(Exception):
    """
    Is raised in case the status code returned from weaviate is
    not handled in the client implementation and suggests an error.
    """
    def __init__(self, message: str, response: Response):
        """
        Is raised in case the status code returned from weaviate is
        not handled in the client implementation and suggests an error.

        Custom code can act on the attributes:
        - status_code
        - json

        Parameters
        ----------
        message: str
            An error message specific to the context, in which the error occurred.
        response: requests.Response
            The request response of which the status code was unexpected.
        """

        # Set error message

        try:
            body = response.json()
        except:
            body = None

        if body is None:
            super().__init__(message)
        else:
            super().__init__(message+"\t"+str(body))

        self.status_code = response.status_code
        self.json = body

    def __str__(self):
        code = str(self.status_code)
        body = str(self.json)
        return f"Unexpected status code: {code}, with response body: {body}"


class ObjectAlreadyExistsException(Exception):
    """
    Object Already Exists Exception.
    """


class AuthenticationFailedException(Exception):
    """
    Authentication Failed Exception.
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class ServerError500Exception(Exception):
    """
    Server Error 500 Exception.
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class SchemaValidationException(Exception):
    """
    Schema Validation Exception.
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message



from typing import Tuple

class ClientConfig:
    """
    Allows to configure the client with additional parameters.
    """

    def __init__(self, timeout_config: Tuple[int, int]=(2, 20)):
        """
        Initialize a ClientConfig class instance.

        Parameters
        ----------
        timeout_config : tuple of int, optional
            Set the timeout config as a tuple of (retries, time out seconds),
            by default (2, 20).
        """

        self.timeout_config = timeout_config



import time
import threading
from typing import Callable
import weaviate
from .submit import SubmitBatches, SubmitBatchesException


class Batcher:
    """
    Manages batches and batch loading. Autocommits both Objects and References
    when the batcher reacher the allocated size. It resets after the batch is 
    loaded to the weaviate, and can be reused.
    """

    def __init__(self,
            client: weaviate.Client,
            batch_size: int=512,
            verbose: bool=False,
            auto_commit_timeout: float=-1.0,
            max_backoff_time: int=300,
            max_request_retries: int=4,
            return_values_callback: Callable=None
        ):
        """
        Initialization of a Batcher class instance.

        Parameters
        ----------
        client : weaviate.Client
            The weaviate client for which to use the Batcher.
        batch_size : int, optional
            Total batcher size, i.e. "objects" batch size + "references" batch size,
            by default 512.
        verbose : bool, optional
            If True the batcher prints statuses of many steps.
            e.g. the return status of the batches,
            by default False.
        auto_commit_timeout : float, optional
            Time, in seconds, the batcher waits before loading all the data,
            no matter how big the batch size. Deactivated when <= 0,
            by default -1.0.
        max_backoff_time : int, optional
            Exponential backoff will never exceed this time in seconds,
            by default 300
        max_request_retries : int, optional
            States how often a request is retried before it is considered as failed,
            by default 4
        return_values_callback : Callable, optional
            If not none this function is called with the return values every time a
            batch is submitted.
            CAUTION if you use this together with the 'auto_commit_timeout'
            you need to ensure thread safety on the callback.
            by default None.
        """

        self._client = client
        # New batches
        self._objects_batch = weaviate.batch.ObjectsBatchRequest()
        self._reference_batch = weaviate.batch.ReferenceBatchRequest()
        # Submission fails are batches that where tried to be submitted but gave errors.
        # They are stored in an extra list with tuples (submission function, batch data)
        # So that they can be submitted on a later moment.
        self._submission_fails = []

        # Configuration
        self._batch_size = batch_size
        self._print_verbose_activated = verbose
        self._submit_batches = SubmitBatches(max_backoff_time, max_request_retries, self._print_verbose_activated)
        self._return_values_callback = return_values_callback

        # Auto commit info
        self._commit_lock: threading.Lock = threading.Lock()
        self._last_update = time.time()
        self._auto_commit_timeout = auto_commit_timeout
        self._auto_commit_watchdog = None
        if self._auto_commit_timeout > 0:
            self._auto_commit_watchdog = AutoCommitWatchdog(self)
            self._auto_commit_watchdog.start()

    def __enter__(self):
        return self

    def update_batches(self) -> None:
        """
        Forces an update of the batches no matter the current batch size.
        Prints errors if there are any.
        """
        with self._commit_lock:
            self._update_batches_force()

    def _update_batch_object(self, create_function, batch_data):
        """ Tries to send the data to the given function.
            Retains the function and data in failed submissions list if not successful.

        :param create_function:
        :param batch_data:
        :return:
        """

        try:
            self._submit_batches.submit_update(create_function, batch_data)
        except SubmitBatchesException:
            print("Error: Object batch was not added after max retries. Will retry with next batch submit")
            self._submission_fails.append((create_function, batch_data))
        else:
            if self._print_verbose_activated:
                print("Updated object batch successfully")

    def _retry_failed_submissions(self):
        """ Tries to resubmit failed submissions
        :return:
        """
        still_failing = []
        for create_func, batch_data in self._submission_fails:
            try:
                self._submit_batches.submit_update(create_func, batch_data)
            except SubmitBatchesException:
                still_failing.append((create_func, batch_data))
        if self._print_verbose_activated:
            if len(self._submission_fails) > 0:
                print("Of", len(self._submission_fails), "/", len(still_failing), "are still failing.")
        self._submission_fails = still_failing

    def _update_batches_force(self):
        if len(self._submission_fails) > 0:
            self._retry_failed_submissions()

        self._update_batch_object(self._client.batch.create, self._objects_batch)
        self._objects_batch = weaviate.batch.ObjectsBatchRequest()

        self._update_batch_object(self._client.batch.create, self._reference_batch)
        self._reference_batch = weaviate.batch.ReferenceBatchRequest()

        result_collection = self._submit_batches.pop_results()
        if self._return_values_callback is not None and len(result_collection) > 0:
            self._return_values_callback(result_collection)

        self._last_update = time.time()

    def _update_batch_if_necessary(self):
        """ Starts a batch load if the batch size is reached
        """
        if len(self._objects_batch) + len(self._reference_batch) >= self._batch_size:
            self._update_batches_force()

    def add_data_object(self, data_object, class_name, uuid=None):
        with self._commit_lock:
            self._last_update = time.time()
            self._objects_batch.add(data_object, class_name, uuid)
            self._update_batch_if_necessary()

    def add_reference(self, from_thing_class_name, from_thing_uuid, from_property_name,
                    to_thing_uuid):
        with self._commit_lock:
            self._last_update = time.time()
            self._reference_batch.add(
                from_entity_class_name=from_thing_class_name,
                from_entity_uuid=from_thing_uuid,
                from_property_name=from_property_name,
                to_entity_uuid=to_thing_uuid
            )
            self._update_batch_if_necessary()

    def close(self):
        """ Closes this Batcher.
            Makes sure that all unfinished batches are loaded into weaviate.
            Batcher is not useable after closing.
        """

        # stop watchdog thread
        if self._auto_commit_watchdog is not None:
            with self._commit_lock:
                self._auto_commit_watchdog.is_closed = True

        retry_counter = 0
        while len(self._objects_batch) > 0 or len(self._reference_batch) > 0 or\
                        len(self._submission_fails) > 0:
            # update batches might have an connection error just retry until it is successful
            self.update_batches()
            retry_counter += 1
            if retry_counter > 500:
                print("CRITICAL ERROR things can not be updated exit after 500 retries")
                exit(5)

        self._reference_batch = None
        self._objects_batch = None
        self._client = None

    def __exit__(self, exception_type, exception_value, traceback):
        try:
            self.close()
        except Exception as e:
            print(e)


class AutoCommitWatchdog(threading.Thread):

    def __init__(self, batcher):
        """

        :param batcher:
        :type batcher: Batcher
        """
        threading.Thread.__init__(self)
        self.batcher = batcher
        self.is_closed = False

    def commit_batcher(self):
        with self.batcher._commit_lock:
            self.batcher._update_batches_force()

    def run(self):
        while not self.is_closed:
            now = time.time()
            delta = now - self.batcher._last_update
            if delta > self.batcher._auto_commit_timeout:
                self.commit_batcher()
            time.sleep(0.125)


from __future__ import unicode_literals
import copy
import warnings
from abc import ABC, abstractmethod
from typing import List
import validators
from weaviate.util import is_weaviate_entity_url, get_uuid_from_weaviate_url

class BatchRequest(ABC):
    """
    BatchRequest abstract class used as a interface for batch requests.
    """

    @abstractmethod
    def get_request_body(self):
        pass

class ReferenceBatchRequest(BatchRequest):
    """
    Collect references to add them in one request to weaviate.
    Caution this request will miss some validations to be faster.
    """

    def __init__(self):
        self._from_entity_class_names = []
        self._from_entity_ids = []
        self._from_entity_properties = []
        self._to_entity_ids = []

    def __len__(self):
        return len(self._from_entity_class_names)

    def add(self,
            from_entity_uuid: str,
            from_entity_class_name: str,
            from_property_name: str,
            to_entity_uuid: str
        ) -> None:
        """
        Add one reference to this batch.

        Parameters
        ----------
        from_entity_uuid : str
            The UUID or URL of the thing that should reference another object.
        from_entity_class_name : str
            The name of the class that should reference another object.
        from_property_name : str
            The name of the property that contains the reference.
        to_entity_uuid : str
            The UUID or URL of the object that is actually referenced.

        Raises
        ------
        TypeError
            If arguments are not of type str.
        """

        if not isinstance(from_entity_class_name, str) or not isinstance(from_entity_uuid, str) or\
            not isinstance(from_property_name, str) or not isinstance(to_entity_uuid, str):
            raise TypeError('All arguments must be of type string')

        if is_weaviate_entity_url(from_entity_uuid):
            from_entity_uuid = get_uuid_from_weaviate_url(from_entity_uuid)
        if is_weaviate_entity_url(to_entity_uuid):
            to_entity_uuid = get_uuid_from_weaviate_url(to_entity_uuid)

        self._from_entity_class_names.append(from_entity_class_name)
        self._from_entity_ids.append(from_entity_uuid)
        self._from_entity_properties.append(from_property_name)
        self._to_entity_ids.append(to_entity_uuid)

    def add_reference(self,
            from_entity_uuid: str,
            from_entity_class_name: str,
            from_property_name: str,
            to_entity_uuid: str
        ) -> None:
        """
        'add_references' is deprecated due to redundancy, use 'add' instead!
        """

        warnings.warn(
            "'add_references' is deprecated due to redundancy, use 'add' instead!",
            DeprecationWarning
        )

        return self.add(
            from_entity_uuid=from_entity_uuid,
            from_entity_class_name=from_entity_class_name,
            from_property_name=from_property_name,
            to_entity_uuid=to_entity_uuid
            )

    def get_request_body(self) -> List[dict]:
        """
        Get request body as a list of dictionaries, where each dictionary
        is a weaviate-schema reference.

        Returns
        -------
        list
            A list of references as dictionaries.
        """

        batch_body = []
        for i in range(len(self)):
            batch_body.append(
                {
                "from": "weaviate://localhost/" + self._from_entity_class_names[i] + "/"
                    + self._from_entity_ids[i] + "/" + self._from_entity_properties[i],
                "to": "weaviate://localhost/" + self._to_entity_ids[i]
                }
            )
        return batch_body


class ObjectsBatchRequest(BatchRequest):
    """
    Collect objects for one batch request to weaviate.
    Caution this batch will not be validated through weaviate.
    """

    def __init__(self):
        self._objects = []

    def __len__(self):
        return len(self._objects)

    def add(self,
            data_object: dict,
            class_name: str,
            uuid: str=None
        ) -> None:
        """
        Add one object to this batch.

        Parameters
        ----------
        data_object : dict
            Object to be added as a dict datatype.
        class_name : str
            The name of the class this object belongs to.
        uuid : str, optional
            UUID of the object as a string, by default None

        Raises
        ------
        TypeError
            If an argument passed is not of an appropriate type.
        ValueError
            If 'uuid' is not of a propper form.
        """

        if not isinstance(data_object, dict):
            raise TypeError("Object must be of type dict")
        if not isinstance(class_name, str):
            raise TypeError("Class name must be of type str")

        batch_item = {
            "class": class_name,
            "properties": copy.deepcopy(data_object)
        }
        if uuid is not None:
            if not isinstance(uuid, str):
                raise TypeError("UUID must be of type str")
            if not validators.uuid(uuid):
                raise ValueError("UUID is not in a proper form")
            batch_item["id"] = uuid

        self._objects.append(batch_item)

    def add_object(self,
            data_object: dict,
            class_name: str,
            uuid: str=None
        ) -> None:
        """
        'add_object' is deprecated due to redundancy, use 'add' instead!
        """

        warnings.warn(
            "'add_object' is deprecated due to redundancy, use 'add' instead!",
            DeprecationWarning
        )

        return self.add(
            data_object=data_object,
            class_name=class_name,
            uuid=uuid
        )

    def get_request_body(self) -> dict:
        """
        Get the request body as it is needed for weaviate

        Returns
        -------
        dict
            The request body as a dict.
        """

        return {
            "fields": [
                "ALL"
            ],
            "objects": self._objects
        }



import sys
import json
from typing import List, Optional
from weaviate.connect import REST_METHOD_POST, Connection
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from .filter import WhereFilter

class AggregateBuilder:
    """
    AggregateBuilder class used to aggregate weaviate objects.
    """

    def __init__(self, class_name: str, connection: Connection):
        """
        Initialize a AggregateBuilder class instance.

        Parameters
        ----------
        class_name : str
            Class name of the objects to be aggregated.
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._class_name = class_name
        self._connection = connection
        self._with_meta_count = False
        self._fields: List[str] = []
        self._where: Optional[WhereFilter] = None
        self._group_by_properties: Optional[List[str]] = None
        self._uses_filter = False

    def with_meta_count(self) -> 'AggregateBuilder':
        """
        Set Meta Count to True.

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._with_meta_count = True
        return self

    def with_fields(self, field: str) -> 'AggregateBuilder':
        """
        Include a field in the aggregate query.

        Parameters
        ----------
        field : str
            Field to include in the aggregate query.
            e.g. '<property_name> { count }'

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._fields.append(field)
        return self

    def with_where(self, filter: dict) -> 'AggregateBuilder':
        """
        Set 'where' filter.

        Parameters
        ----------
        filter : dict
            The where filter to include in the aggregate query.

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._where = WhereFilter(filter)
        self._uses_filter = True
        return self

    def with_group_by_filter(self, properties: List[str]) -> 'AggregateBuilder':
        """
        Add a group by filter to the query. Might requires the user to set
        an additional group by clause using `with_fields(..)`.

        Parameters
        ----------
        properties : list of str
            list of properties that are included in the group by filter.
            Generates a filter like: 'groupBy: ["property1", "property2"]'
            from a list ["property1", "property2"]

        Returns
        -------
        weaviate.gql.aggregate.AggregateBuilder
            Updated AggregateBuilder.
        """

        self._group_by_properties = properties
        self._uses_filter = True
        return self

    def build(self) -> str:
        """
        Build the query and return the string.

        Returns
        -------
        str
            The GraphQL query as a string.
        """

        # Path
        query = f"{{Aggregate{{{self._class_name}"

        # Filter
        if self._uses_filter:
            query += "("
        if self._where is not None:
            query += str(self._where)
        if self._group_by_properties is not None:
            query += f"groupBy: {json.dumps(self._group_by_properties)}"
        if self._uses_filter:
            query += ")"

        # Body
        query += "{"
        if self._with_meta_count:
            query += "meta{count}"
        for field in self._fields:
            query += field

        # close
        query += "}}}"
        return query

    def do(self) -> dict:
        """
        Builds and runs the query.

        Returns
        -------
        dict
            The response of the query.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        query = self.build()

        try:
            response = self._connection.run_rest("/graphql", REST_METHOD_POST, {"query": query})
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, query was not successful.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 200:
            return response.json()  # success
        raise UnexpectedStatusCodeException("Query was not successful", response)



import sys
from typing import Union, List
from weaviate.connect import REST_METHOD_DELETE, REST_METHOD_PUT, REST_METHOD_POST, Connection
from weaviate.exceptions import RequestsConnectionError, UnexpectedStatusCodeException
from weaviate.util import is_valid_uuid



class Reference:
    """
    Reference class used to manipulate refferences within objects.
    """

    # This is a class private variable that contains implemented
    # methods and their respective REST method and Success return code.
    _methods_to_rest = {
        "add" : {
            "rest" : REST_METHOD_POST,
            "success_code": 200,
        },
        "update" : {
            "rest" : REST_METHOD_PUT,
            "success_code" : 200,
        },
        "delete" : {
            "rest" : REST_METHOD_DELETE,
            "success_code" : 204
        }
    }

    def __init__(self, connection: Connection):
        """
        Initialize a Reference class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    def delete(self,
            from_uuid: str,
            from_property_name: str,
            to_uuid: str
        ) -> None:
        """
        Remove a reference to another thing. Equal to removing one
        direction of an edge from the graph.

        Parameters
        ----------
        from_uuid : str
            The ID of the object that references another object.
        from_property_name : str
            The property from which the reference should be deleted.
        to_uuid : str
            The UUID of the referenced object.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If parameter has the wrong type.
        ValueError
            If uuid is not properly formed.
        """

        self._try_run_rest(
            from_uuid=from_uuid,
            from_property_name=from_property_name,
            to_uuid_s=to_uuid,
            method="delete"
        )

    def update(self,
            from_uuid: str,
            from_property_name: str,
            to_uuids: list
        ) -> None:
        """
        Allows to update all references in that property with a new set of references.
        All old references will be replaced.

        Parameters
        ----------
        from_uuid : str
            The object that should have the reference as part of its properties.
            Should be in the form of an UUID or in form of an URL.
            E.g.
            'http://localhost:8080/v1/objects/fc7eb129-f138-457f-b727-1b29db191a67'
            or
            'fc7eb129-f138-457f-b727-1b29db191a67'
        from_property_name : str
            The name of the property within the object.
        to_uuids : list
            The UUIDs of the objects that should be referenced.
            Should be a list of str in the form of an UUID or str in form of an URL.
            E.g.
            ['http://localhost:8080/v1/objects/fc7eb129-f138-457f-b727-1b29db191a67', ...]
            or
            ['fc7eb129-f138-457f-b727-1b29db191a67', ...]

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If the parameters are of the wrong type.
        ValueError
            If the parameters are of the wrong value.
        """

        self._try_run_rest(
            from_uuid=from_uuid,
            from_property_name=from_property_name,
            to_uuid_s=to_uuids,
            method="update"
        )

    def add(self,
            from_uuid: str,
            from_property_name: str,
            to_uuid: str
        ) -> None:
        """
        Allows to link an object to an object unidirectionally.

        Parameters
        ----------
        from_uuid : str
            The ID of the object that should have the reference as part
            of its properties. Should be a plane UUID or an URL.
            E.g.
            'http://localhost:8080/v1/objects/fc7eb129-f138-457f-b727-1b29db191a67'
            or
            'fc7eb129-f138-457f-b727-1b29db191a67'
        from_property_name : str
            The name of the property within the object.
        to_uuid : str
            The UUID of the object that should be referenced.
            Should be a plane UUID or an URL.
            E.g.
            'http://localhost:8080/v1/objects/fc7eb129-f138-457f-b727-1b29db191a67'
            or
            'fc7eb129-f138-457f-b727-1b29db191a67'

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If the parameters are of the wrong type.
        ValueError
            If the parameters are of the wrong value.
        """

        self._try_run_rest(
            from_uuid=from_uuid,
            from_property_name=from_property_name,
            to_uuid_s=to_uuid,
            method="add"
        )

    def _try_run_rest(self,
            from_uuid: str,
            from_property_name: str,
            to_uuid_s: Union[str, List[str]],
            method: str
        ) -> None:
        """
        Try to run a rest method based on the 'method' argument.

        Parameters
        ----------
        from_uuid : str
            The ID of the object that should have the reference as part
            of its properties. Should be a plane UUID or an URL.
        from_property_name : str
            The name of the property within the object.
        to_uuid_s : Union[str, List[str]]
            The UUID/s of the object/s that should be referenced.
            It can be a string or a list of strings. List of string is only
            used for the 'update' method.
        method : str
            REST call method, check self._methods_to_rest to see what methods
            are available at the moment.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If the parameters are of the wrong type.
        ValueError
            If the parameters are of the wrong value.
        """

        if method not in self._methods_to_rest.keys():
            raise ValueError(f"'{method}' not supported!")

        # Validate and create Beacon
        _validate_uuid(from_uuid)
        _validate_property_name(from_property_name)

        # Validate 'to_uuid_s' and create the respective beacon/s
        beacons = []
        if method == "update":
            # Update works with a list of 'to_uuids' so it needs a separate check.
            if not isinstance(to_uuid_s, list):
                to_uuid_s = [to_uuid_s]

            for to_uuid in to_uuid_s:
                _validate_uuid(to_uuid)
                beacons.append(_get_beacon(to_uuid))
        else:
            # Other methods take just a UUID as a str
            _validate_uuid(to_uuid_s)
            beacons = _get_beacon(to_uuid_s)

        # Try to Run REST method
        path = f"/objects/{from_uuid}/references/{from_property_name}"
        try:
            response = self._connection.run_rest(
                path=path,
                rest_method=self._methods_to_rest[method]["rest"],
                weaviate_object=beacons
                )
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                    + f' Connection error, did not {method} reference.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])

        if response.status_code == self._methods_to_rest[method]["success_code"]:
            return
        raise UnexpectedStatusCodeException(f"{method} property reference to object", response)


def _get_beacon(to_uuid: str) -> dict:
    """
    Get a weaviate-style beacon.

    Parameters
    ----------
    to_uuid : str
        The UUID to create beacon for.

    Returns
    -------
    dict
        Weaviate-style beacon as a dict.
    """

    return {
        "beacon": f"weaviate://localhost/{to_uuid}"
    }


def _validate_property_name(property_name: str) -> None:
    """
    Validate the property name.

    Parameters
    ----------
    property_name : str
        Property name to be validated.

    Raises
    ------
    TypeError
        If 'property_name' is not of type str.
    """

    if not isinstance(property_name, str):
        raise TypeError("from_property_name must be of type 'str' but was: "\
                        + str(type(property_name)))


def _validate_uuid(uuid: str) -> None:
    """
    Validate the UUID.

    Parameters
    ----------
    uuid : str
        The UUID to be validated.

    Raises
    ------
    ValueError
        If 'uuid' is not valid or UUID can not be extracted
        from 'uuid'.
    """

    if not is_valid_uuid(uuid):
        raise ValueError("Not valid uuid or uuid can not be extracted from value")



from .auth import AuthClientCredentials, AuthClientPassword
from .client_config import ClientConfig
from .exceptions import *
from .util import generate_local_beacon
from .batch import ReferenceBatchRequest, ObjectsBatchRequest
from .classification import Classification
from .client import Client



import json
import os
from typing import Union
import validators
import requests


def generate_local_beacon(to_uuid: str) -> dict:
    """
    Generates a beacon with the given uuid.

    Parameters
    ----------
    to_uuid : str
        The UUID for which to create a local beacon.

    Returns
    -------
    dict
        The local beacon.

    Raises
    ------
    TypeError
        If 'to_uuid' is not of type str.
    ValueError
        If the 'to_uuid' is not valid.
    """

    if not isinstance(to_uuid, str):
        raise TypeError("Expected to_thing_uuid of type str")
    if not validators.uuid(to_uuid):
        raise ValueError("Uuid does not have the propper form")
    return {"beacon": "weaviate://localhost/" + to_uuid}


def _get_dict_from_object(object_: Union[str, dict]) -> dict:
    """
    Takes an object that should describe a dict
    e.g. a schema or a object and tries to retrieve the dict.

    Parameters
    ----------
    object_ : str or dict
        The object from which to retrieve the dict.
        Can be a python dict, or the path to a json file or a url of a json file.

    Returns
    -------
    dict
        The object as a dict.

    Raises
    ------
    TypeError
        If neither a string nor a dict.
    ValueError
        If no dict can be retrieved from object.
    """

    # check if things files is url
    if object_ is None:
        raise TypeError("argument is None")

    if isinstance(object_, dict):
        # Object is already a dict
        return object_
    if isinstance(object_, str):
        if validators.url(object_):
            # Object is URL
            response = requests.get(object_)
            if response.status_code == 200:
                return response.json()
            raise ValueError("Could not download file " + object_)

        if not os.path.isfile(object_):
            # Object is neither file nor URL
            raise ValueError("No file found at location " + object_)
        # Object is file
        with open(object_, 'r') as file:
            return json.load(file)
    raise TypeError("Argument is not of the supported types. Supported types are \
                                    url or file path as string or schema as dict.")


def is_weaviate_entity_url(url: str) -> bool:
    """
    Checks if the input follows a normal url like this:
    'weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'

    Parameters
    ----------
    input : str
        The URL to be validated.

    Returns
    -------
    bool
        True if 'input' is an weaviate entity URL.
        False otherwise.
    """

    if not isinstance(url, str):
        return False
    if not url.startswith("weaviate://"):
        return False
    url = url[11:]
    split = url.split("/")
    if len(split) != 2:
        return False
    if split[0] != "localhost":
        if not validators.domain(split[0]):
            return False
    if not validators.uuid(split[1]):
        return False
    return True


def is_object_url(url: str) -> bool:
    """
    Validates if an url like http://localhost:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446
    references a object. It only validates the path not the host or the protocol.

    Parameters
    ----------
    input : str
        The URL to be validated.

    Returns
    -------
    bool
        True if the 'input' is a valid path to an object.
        False otherwise.
    """

    split = url.split("/")
    if len(split) < 3:
        return False
    if not validators.uuid(split[-1]):
        return False
    if not split[-2] == "objects":
        return False
    if not split[-3] == "v1":
        return False
    return True


def is_valid_uuid(uuid: str) -> bool:
    """
    Validate the UUID.

    Parameters
    ----------
    uuid : str
        The UUID to be validated.
        Should be in the form of an UUID or in form of an URL.
        E.g.
        'http://localhost:8080/v1/objects/fc7eb129-f138-457f-b727-1b29db191a67'
        or
        'fc7eb129-f138-457f-b727-1b29db191a67'

    Raises
    ------
    TypeError
        If 'uuid' is not of type str.
    """

    if not isinstance(uuid, str):
        raise TypeError("uuid must be of type str but was: " + str(type(uuid)))

    _is_weaviate_url = is_weaviate_entity_url(uuid)
    _is_object_url = is_object_url(uuid)
    _uuid = uuid
    if _is_weaviate_url or _is_object_url:
        split = uuid.split("/")
        _uuid = split[-1]
    return validators.uuid(_uuid)


def get_uuid_from_weaviate_url(url: str) -> str:
    """
    Get the UUID from a weaviate URL.

    Parameters
    ----------
    url : str
        The weaviate URL.
        Of this form: 'weaviate://localhost/objects/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'

    Returns
    -------
    str
        The UUID.
    """

    return url.split("/")[-1]


def get_domain_from_weaviate_url(url: str) -> str:
    """
    Get the domain from a weaviate URL.

    Parameters
    ----------
    url : str
        The weaviate URL.
        Of this form: 'weaviate://localhost/objects/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'

    Returns
    -------
    str
        The domain.
    """

    return url[11:].split("/")[0]


def _is_sub_schema(sub_schema: dict, schema: dict) -> bool:
    """
    Check for a subset in a schema.

    Parameters
    ----------
    sub_schema : dict
        The smaller schema that should be contained in the 'schema'.
    schema : dict
        The schema for which to check if 'sub_schema' is a part of.

    Returns
    -------
    bool
        True is 'sub_schema' is a subset of the 'schema'.
        False otherwise.
    """

    schema_classes = schema["objects"].get("classes", [])
    sub_schema_classes = sub_schema["objects"].get("classes", [])
    return _compare_class_sets(sub_schema_classes, schema_classes)


def _compare_class_sets(sub_set: list, set_: list) -> bool:
    """
    Check for a subset in a set of classes.

    Parameters
    ----------
    sub_set : list
        The smaller set that should be contained in the 'set'.
    schema : dict
        The set for which to check if 'sub_set' is a part of.

    Returns
    -------
    bool
        True is 'sub_set' is a subset of the 'set'.
        False otherwise.
    """

    for sub_set_class in sub_set:
        found = False
        for set_class in set_:
            if sub_set_class["class"] == set_class["class"]:
                if _compare_properties(sub_set_class["properties"], set_class["properties"]):
                    found = True
                    break
        if not found:
            return False
    return True


def _compare_properties(sub_set: list, set_: list) -> bool:
    """
    Check for a subset in a set of properties.

    Parameters
    ----------
    sub_set : list
        The smaller set that should be contained in the 'set'.
    schema : dict
        The set for which to check if 'sub_set' is a part of.

    Returns
    -------
    bool
        True is 'sub_set' is a subset of the 'set'.
        False otherwise.
    """

    for sub_set_property in sub_set:
        found = False
        for set_property in set_:
            if sub_set_property["name"] == set_property["name"]:
                found = True
                break
        if not found:
            return False
    return True



"""
Schema validation module.
"""
from typing import Any
from weaviate.exceptions import SchemaValidationException


def validate_schema(schema: dict) -> None:
    """
    Validate schema.

    Parameters
    ----------
    schema : dict
        Schema to be validated.

    Raises
    ------
    SchemaValidationException
        If the schema could not be validated against the standard format.
    """

    # check if schema has only "classes" as keys
    if len(schema) != 1 or "classes" not in schema:
        raise SchemaValidationException('each schema has to have only "classes" \
                    in the first level of the JSON format file/parameter/object')
    # check if "classes" is of type list
    _check_key_type("classes", schema["classes"], list)
    # check if each class in the "classes" is a valid class
    for weaviate_class in schema["classes"]:
        check_class(weaviate_class)


def check_class(class_definition: dict) -> None:
    """
    Validate a class against the standard class format.

    Parameters
    ----------
    class_definition : dict
        The definition of the class to be validated.

    Raises
    ------
    SchemaValidationException
        If the class could not be validated against the standard class format.
    """

    # check mandatory keys
    if "class" not in class_definition:
        raise SchemaValidationException('"class" key is missing in class definition.')

    # check optional keys
    for key in class_definition.keys():
        # Check if key is known
        if key not in ["class", "vectorIndexType", "vectorIndexConfig", "moduleConfig",\
            "description", "vectorizer", "properties"]:
            raise SchemaValidationException(f'"{key}" is not a known class definition key.')
        # check if key is right type
        if key in ["class", "vectorIndexType", "description", "vectorizer"]:
            _check_key_type(key, class_definition[key], str)
        if key in ["vectorIndexConfig", "moduleConfig"]:
            _check_key_type(key, class_definition[key], dict)
        if key in ["properties"]:
            _check_key_type(key, class_definition[key], list)
        # TODO check in depth dicts

    if "properties" in class_definition:
        for class_property in class_definition["properties"]:
            check_property(class_property)


def check_property(class_property: dict) -> None:
    """
    Validate a class property against the standard class property.

    Parameters
    ----------
    class_property : dict
        The class property to be validated.

    Raises
    ------
    SchemaValidationException
        If the class property could not be validated against\
        the standard class property format.
    """

    # mandatory fields
    if "dataType" not in class_property:
        raise SchemaValidationException('Property does not contain "dataType"')
    if "name" not in class_property:
        raise SchemaValidationException('Property does not contain "name"')

    for key in class_property:
        # check for misspelled and/or non-existent properties
        if key not in ["dataType", "name", "moduleConfig", "description", "indexInverted"]:
            raise SchemaValidationException(f'Property "{key}" is not known.')

        # Test types
        if key in ["dataType"]:
            _check_key_type(key, class_property[key], list)
        if key in ["name", "description"]:
            _check_key_type(key, class_property[key], str)
        if key in ["indexInverted"]:
            _check_key_type(key, class_property[key], bool)
        if key in ["moduleConfig"]:
            _check_key_type(key, class_property[key], dict)
            # TODO check "moduleConfig" hierarch types and keys

    # Test dataType types
    for data_type in class_property["dataType"]:
        _check_key_type("dataType", data_type, str)


def _check_key_type(key: str, value: Any, expected_type: Any) -> None:
    """
    Check if value is of an expected type.

    Parameters
    ----------
    key : str
        The key for which to check data type.
    value : Any
        The value of the 'key' for which to check data type.
    expected_type : Any
        The expected data type of the 'value'.

    Raises
    ------
    SchemaValidationException
        If the 'value' is of wrong data type.
    """

    if not isinstance(value, expected_type):
        raise SchemaValidationException(f'"{key}" is type {type(value)} \
                                        but should be {expected_type}.')



import json
from typing import Union, List, Optional


class NearText:
    """
    NearText class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a NearText class instance.

        Parameters
        ----------
        content : dict
            The content of the `nearText` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content'  has key "certainty" but the value is not float.
        """

        if not isinstance(content, dict):
            raise TypeError(f"{self.__class__.__name__} filter is expected to \
                                        be type dict but was {type(content)}")

        self.concepts = _check_concept(content)
        self.certainty: Optional[float] = None
        self.move_to: Optional[dict] = None
        self.move_away_from: Optional[dict] = None

        if "certainty" in content:
            if not isinstance(content["certainty"], float):
                raise TypeError(f"certainty is expected to be a float but was \
                                {type(content['certainty'])}")

            self.certainty = content["certainty"]

        if "moveTo" in content:
            self.move_to = _check_direction_clause(content["moveTo"])

        if "moveAwayFrom" in content:
            self.move_away_from = _check_direction_clause(content["moveAwayFrom"])

    def __str__(self):
        near_text = f'nearText: {{concepts: {json.dumps(self.concepts)} '
        if self.certainty is not None:
            near_text += f'certainty: {str(self.certainty)} '
        if self.move_to is not None:
            near_text += f'moveTo:{{concepts: {json.dumps(self.move_to["concepts"])} \
                        force: {self.move_to["force"]}}} '
        if self.move_away_from is not None:
            near_text += f'moveAwayFrom:{{concepts: {json.dumps(self.move_away_from["concepts"])} \
                        force: {self.move_away_from["force"]}}} '
        return near_text + '} '


class NearVector:
    """
    NearVector class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a NearVector class instance.

        Parameters
        ----------
        content : list
            The content of the `nearVector` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If 'content' does not contain "vector".
        TypeError
            If 'content["vector"]' is not of type list.
        AttributeError
            If invalid 'content' keys are provided.
        """

        if not isinstance(content, dict):
            raise TypeError(f"{self.__class__.__name__} filter is expected to \
                be type dict but was {type(content)}")

        self.vector = _check_vector(content)
        self.certainty: Optional[float] = None

        # Check optional fields

        if "certainty" in content:
            if not isinstance(content["certainty"], float):
                raise TypeError(f"certainty is expected to be a float but was \
                            {type(content['certainty'])}")
            self.certainty = content["certainty"]

    def __str__(self):
        near_vector = f'nearVector: {{vector: {json.dumps(self.vector)}'
        if self.certainty is not None:
            near_vector += f' certainty: {self.certainty}'
        return near_vector + '} '


class WhereFilter:
    """
    WhereFilter class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a WhereFilter class instance.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If a mandatory key is missing in the filter content.
        """

        if not isinstance(content, dict):
            raise TypeError(f"WhereFilter is expected to be type dict but was {type(content)}")

        if "path" in content:
            self.is_filter = True
            self._parse_filter(content)
        elif "operands" in content:
            self.is_filter = False
            self._parse_operator(content)
        else:
            raise ValueError("Filter is missing required fileds: ", content)

    def _parse_filter(self, content: dict) -> None:
        """
        Set filter fields for the WhereFilter.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        ValueError
            If 'content' is missing required fields.
        """

        if "operator" not in content:
            raise ValueError("Filter is missing required fileds: ", content)

        self.path = json.dumps(content["path"])
        self.operator = content["operator"]
        self.value_type = _find_value_type(content)
        self.value = content[self.value_type]

    def _parse_operator(self, content: dict) -> None:
        """
        Set operator fields for the WhereFilter.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        ValueError
            If 'content' is missing required fields.
        """

        if "operator" not in content:
            raise ValueError("Filter is missing required fileds: ", content)

        self.operator = content["operator"]
        self.operands = []
        for operand in content["operands"]:
            self.operands.append(WhereFilter(operand))

    def __str__(self):
        if self.is_filter:
            gql = f'where: {{path: {self.path} operator: {self.operator} {self.value_type}: '
            if self.value_type in ["valueInt", "valueNumber"]:
                gql += f'{self.value}}}'
            elif self.value_type == "valueBoolean":
                bool_value = str(self.value).lower()
                gql += f'{bool_value}}}'
            elif self.value_type == "valueGeoRange":
                geo_value = json.dumps(self.value)
                gql += f'{geo_value}}}'
            else:
                gql += f'"{self.value}"}}'
            return gql + ' '

        operands_str = []
        for operand in self.operands:
            operands_str.append(str(operand))
        operands = ", ".join(operands_str)
        return f'where: {{operator: {self.operator} operands: [{operands}]}} '


def _check_direction_clause(direction: dict) -> dict:
    """
    Validate the direction sub clause.

    Parameters
    ----------
    direction : dict
        A sub clause of the Explore filter.

    Returns
    -------
    dict
        Returns back the original 'direction' if it passed the validation.

    Raises
    ------
    TypeError
        If 'direction' is not a dict.
    TypeError
        If the value of the "force" key is not float.
    ValueError
        If no "force" key in the 'direction'.
    """

    if not isinstance(direction, dict):
        raise TypeError(f"move clause should be dict but was {type(direction)}")
    _check_concept(direction)
    if not "force" in direction:
        raise ValueError("move clause needs to state a force")
    if not isinstance(direction["force"], float):
        raise TypeError(f"force should be float but was {type(direction['force'])}")
    return direction


def _check_concept(content: dict) -> Union[list, str]:
    """
    Validate the concept sub clause.

    Parameters
    ----------
    content : dict
        An Explore (sub) clause to check for 'concepts'.

    Returns
    -------
    list or str
        Concept/s of the (sub) clause.

    Raises
    ------
    ValueError
        If no "concepts" key in the 'content' dict.
    TypeError
        If the value of the  "concepts" is of wrong type.
    """

    if "concepts" not in content:
        raise ValueError("No concepts in content")

    if not isinstance(content["concepts"], (list, str)):
        raise TypeError(f"Concepts must be of type list or str not {type(content['concepts'])}")
    return content["concepts"]


def _check_vector(content: dict) -> List[float]:
    """
    Validate the vector of the nearVector.

    Parameters
    ----------
    content : dict
        A nearVector clause to validate.

    Returns
    -------
    list or str
        The vector of the clause.

    Raises
    ------
    ValueError
        If no "vector" key in the 'content' dict.
    TypeError
        If the value of the  "vector" is of wrong type.
    """

    if "vector" not in content:
        raise ValueError("No 'vector' in 'content'.")
    if not isinstance(content["vector"], list):
        raise TypeError(f"'vector' key is expected to be type list but was {content['vector']}")
    return content["vector"]


def _find_value_type(content: dict) -> str:
    """
    Find the correct type of the content.

    Parameters
    ----------
    content : dict
        The content for which to find the appropriate data type.

    Returns
    -------
    str
        The correct data type.

    Raises
    ------
    ValueError
        If missing required fields.
    """

    if "valueString" in content:
        return "valueString"
    if "valueText" in content:
        return "valueText"
    if "valueInt" in content:
        return "valueInt"
    if "valueNumber" in content:
        return "valueNumber"
    if "valueDate" in content:
        return "valueDate"
    if "valueBoolean" in content:
        return "valueBoolean"
    if "valueGeoRange" in content:
        return "valueGeoRange"
    raise ValueError("Filter is missing required fileds: ", content)



import base64
import copy
from abc import ABC, abstractmethod

class AuthCredentials(ABC):
    """
    Base class for getting the grand type and credentials.
    """

    def __init__(self):
        self._credentials_body = {}

    @abstractmethod
    def get_credentials(self) -> dict:
        """
        Get credentials.
        """

class AuthClientCredentials(AuthCredentials):
    """
    Using a client secret for authentication.
    In case of grand type client credentials.
    """

    def __init__(self, client_secret: str):
        """
        Using a client secret for authentication.
        In case of grand type client credentials.

        Parameters
        ----------
        client_secret : str
            The access token.
        """

        super().__init__()
        self._credentials_body["grant_type"] = "client_credentials"
        self.client_secret_encoded = base64.b64encode(client_secret.encode('ascii')).decode('ascii')

    def get_credentials(self) -> dict:
        """
        Get decoded credentials.

        Returns
        -------
        dict
            Decoded credentials.
        """

        return_body = copy.deepcopy(self._credentials_body)
        return_body["client_secret"] = (
            base64.b64decode(self.client_secret_encoded.encode('ascii'))
            .decode('ascii')
        )
        return return_body


class AuthClientPassword(AuthCredentials):
    """
    Using username and password for authentication.
    In case of grand type password.
    """

    def __init__(self, username: str, password: str) -> None:
        """
        Using username and password for authentication.
        In case of grand type password.

        Parameters
        ----------
        username : str
            The username to login with.
        password : str
            Password fot the given User.
        """

        super().__init__()
        self._credentials_body["grant_type"] = "password"
        self._credentials_body["username"] = username
        self.password_encoded = base64.b64encode(password.encode('ascii')).decode('ascii')

    def get_credentials(self) -> dict:
        """
        Get decoded credentials.

        Returns
        -------
        dict
            Decoded credentials.
        """

        return_body = copy.deepcopy(self._credentials_body)
        return_body["password"] = (
            base64.b64decode(self.password_encoded.encode('ascii'))
            .decode('ascii')
        )
        return return_body



import time
from requests.exceptions import ReadTimeout, Timeout
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError


class SubmitBatchesException(Exception):
    def __init__(self, message):
        super().__init__(message)


class SubmitBatches:

    def __init__(self,
            max_backoff_time: int = 300,
            max_request_retries: int = 4,
            verbose_enabled: bool = False
        ):
        """
        Initialize a SubmitBatches class instance.

        Parameters
        ----------
        max_backoff_time : int, optional
            Max time used in the exponential backoff, by default 300.
        max_request_retries : int, optional
            States how often a request is retried before it counts as failed, by default 4.
        verbose_enabled : bool, optional
            If True errors will be printed directly to stdout, by default False.
        """

        self._backoff_time = 10
        self._backoff_count = 0
        self._max_backoff_time = max_backoff_time
        self._result_collection = []
        self._verbose_enabled = verbose_enabled
        self._max_request_retries = max_request_retries

    def pop_results(self) -> list:
        """
        Pops the results of the submissions.

        Returns
        -------
        list
            The results from all submitted updates since the last pop.
        """

        return_value = self._result_collection
        self._result_collection = []
        return return_value

    def submit_update(self, create_func, data) -> None:
        """

        :param create_func: the function that should be called with the data
        :param data: the data that should be used as parameter to the create_func
        :return:
        :raises:
            SubmitBatchesException if the batch could not be submitted
        """

        if len(data) == 0:
            return

        retry_counter = 0
        while not self._submit_data(create_func, data):
            self._sleep_backoff()
            retry_counter += 1
            if retry_counter >= self._max_request_retries:
                raise SubmitBatchesException("Could not submit data")

        self._reset_backoff()

    def _submit_data(self, create_func, data):
        """ Submits data using the create func and appends the result to the result collection

        :param create_func: callback function
        :param data: parameter for callback function
        :return: False if failed, True if successful
        """
        try:
            result = create_func(data)
        except (RequestsConnectionError, Timeout, ReadTimeout, UnexpectedStatusCodeException) as e:
            print("Exception in creating data: ", e)
            return False
        else:
            if isinstance(result, list):
                self._result_collection += result
                self._print_errors(result)
            return True

    def _sleep_backoff(self):
        """ Calculates an exponential backoff and sleeps for the calculated time.
            Is limited by the max_backoff_time settable in the constructor
        :return:
        """
        self._backoff_time = self._backoff_time + self._backoff_count * self._backoff_time
        if self._backoff_time > self._max_backoff_time:
            self._backoff_time = self._max_backoff_time
        else:
            self._backoff_count += 1
        time.sleep(self._backoff_time)

    def _reset_backoff(self):
        self._backoff_time = 10
        self._backoff_count = 0

    def _print_errors(self, request_result):
        if not self._verbose_enabled:
            return

        for item in request_result:
            if len(item['result']) > 0:
                if not 'status' in item['result']:
                    print(item)
                    return

                if item['result']['status'] != 'SUCCESS':
                    print(item)



import uuid
from typing import Any
import rdflib
import weaviate
import weaviate.tools as tools
from weaviate.rdf.constants import triple_schema


class TripleLoader:

    def __init__(self, client: weaviate.Client):
        """
        Initialize a TripleLoader class instance.

        Parameters
        ----------
        client : weaviate.Client
            An active instance of weaviate client.
        """

        self._client = client
        # To avoid adding too many of the same objects we will cache their uuids
        # This is not strictly necessary just an optimization
        self._uuid_cache = {}
        self._batcher = None

    def _add_schema(self) -> None:
        """
        Add schema to weaviate.
        """
        if not self._client.schema.contains(triple_schema):
            self._client.schema.create(triple_schema)

    def _add_rdf_object(self, value: Any, class_name: str) -> str:
        """
        Create the object in weaviate.

        Parameters
        ----------
        value: any
            Value of the object.
        class_name : str
            Either Subject, Predicate or Object.

        Returns
        -------
        str
            The UUID of the object.
        """

        object_id = tools.generate_uuid(value, class_name)
        if object_id in self._uuid_cache:
            return object_id
        self._uuid_cache[object_id] = True

        object_ = {
            "value": str(value),
            "valueKey": str(value)
        }
        self._batcher.add_data_object(object_, class_name, object_id)
        return object_id

    def add_graph(self, graph: rdflib.Graph) -> None:
        """
        Add RDF graph to weaviate.

        Parameters
        ----------
        graph : rdflib.Graph
            The RDF graph to be added.
        """

        self._add_schema()

        self._batcher = tools.Batcher(self._client)

        for s, p, o in graph:
            s_id = self._add_rdf_object(s, "Subject")
            p_id = self._add_rdf_object(p, "Predicate")
            o_id = self._add_rdf_object(o, "Object")

            t_id = str(uuid.uuid4())
            self._batcher.add_data_object({}, "Triple", t_id)
            self._batcher.add_reference("Triple", t_id, "subject", s_id)
            self._batcher.add_reference("Triple", t_id, "predicate", p_id)
            self._batcher.add_reference("Triple", t_id, "object", o_id)
        self._batcher.close()



import sys
from typing import List, Union
from weaviate.connect import REST_METHOD_POST, Connection
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from .get import GetBuilder
from .aggregate import AggregateBuilder


class Query:
    """
    Query class used to make `get` and/or `aggregate` GraphQL queries.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Classification class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    def get(self,
            class_name: str,
            properties: Union[List[str], str]
        ) -> GetBuilder:
        """
        Instantiate a GetBuilder for GraphQL `get` requests.

        Parameters
        ----------
        class_name : str
            Class name of the objects to interact with.
        properties : list of str or str
            Properties of the objetcs to interact with.

        Returns
        -------
        GetBuilder
            A GetBuilder to make GraphQL `get` requests from weaviate.
        """

        return GetBuilder(class_name, properties, self._connection)

    def aggregate(self, class_name: str) -> AggregateBuilder:
        """
        Instantiate an AggregateBuilder for GraphQL `aggregate` requests.

        Parameters
        ----------
        class_name : str
            Class name of the objects to be aggregated.

        Returns
        -------
        AggregateBuilder
            An AggregateBuilder to make GraphQL `aggregate` requests from weaviate.
        """

        return AggregateBuilder(class_name, self._connection)

    def raw(self, gql_query: str) -> dict:
        """
        Allows to send simple graph QL string queries.
        Be cautious of injection risks when generating query strings.

        Parameters
        ----------
        gql_query : str
            GraphQL query as a string.

        Returns
        -------
        dict
            Data response of the query.

        Raises
        ------
        TypeError
            If 'gql_query' is not of type str.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if not isinstance(gql_query, str):
            raise TypeError("Query is expected to be a string")

        json_query = {"query": gql_query}

        try:
            response = self._connection.run_rest("/graphql", REST_METHOD_POST, json_query)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, query not executed.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])

        if response.status_code == 200:
            return response.json()  # Successfully queried
        raise UnexpectedStatusCodeException("GQL query", response)



import datetime
import time


def get_epoch_time() -> int:
    """
    Get the current epoch time as an integer.

    Returns
    -------
    int
        Current epoch time.
    """

    dts = datetime.datetime.utcnow()
    return round(time.mktime(dts.timetuple()) + dts.microsecond / 1e6)



"""
Module used to automaticaly submit batches to weaviate.
"""
from .batcher import Batcher
import uuid


def generate_uuid(identifier, namespace=""):
    """ Generate a uuid, may be used to consistently generate
    the same UUID for a specific identifier and namespace

    :param identifier: that should be used as basis for the uuid
    :param namespace: allows to namespace the identifier
    :return: uuid
    :rtype: str
    """
    return str(uuid.uuid5(uuid.NAMESPACE_DNS, str(namespace) + str(identifier)))



"""
Triple schema as constant.
"""

triple_schema = {
  "objects": {
    "type": "thing",
    "name": "rdf-triples",
    "classes": [
      {
        "class": "Subject",
        "description": "Subject",
        "keywords": [],
        "properties": [
          {
            "name": "value",
            "description": "value of subject",
            "dataType": [
              "text"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          },
          {
            "name": "valueKey",
            "description": "valueKey of subject",
            "dataType": [
              "string"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          }
        ]
      },
      {
        "class": "Predicate",
        "description": "predicate",
        "keywords": [],
        "properties": [
          {
            "name": "value",
            "description": "value",
            "dataType": [
              "text"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          },
          {
            "name": "valueKey",
            "description": "valueKey",
            "dataType": [
              "string"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          }
        ]
      },
      {
        "class": "Object",
        "description": "object",
        "keywords": [],
        "properties": [
          {
            "name": "value",
            "description": "value",
            "dataType": [
              "text"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          },
          {
            "name": "valueKey",
            "description": "valueKey of subject",
            "dataType": [
              "string"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          }
        ]
      },
      {
        "class": "Triple",
        "description": "rdf triple",
        "keywords": [],
        "properties": [
          {
            "name": "subject",
            "description": "subject of the triple",
            "dataType": [
              "Subject"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          },
          {
            "name": "predicate",
            "description": "predicate of the triple",
            "dataType": [
              "Predicate"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          },
          {
            "name": "object",
            "description": "object of the triple",
            "dataType": [
              "Object"
            ],
            "cardinality": "atMostOne",
            "keywords": []
          }
        ]
      }
    ]
  }
}



import sys
from typing import List, Union, Optional
from weaviate.gql.filter import WhereFilter, NearText, NearVector
from weaviate.connect import REST_METHOD_POST, Connection
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError


class GetBuilder:
    """
    GetBuilder class used to create GraphQL queries.
    """

    def __init__(self,
            class_name: str,
            properties: Union[List[str], str],
            connection: Connection
        ):
        """
        Initialize a GetBuilder class instance.

        Parameters
        ----------
        class_name : str
            Class name of the objects to interact with.
        properties : list of str or str
            Properties of the objetcs to interact with.
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.

        Raises
        ------
        TypeError
            If argument/s is/are of wrong type.
        """
        self._connection = connection

        if not isinstance(class_name, str):
            raise TypeError(f"class name must be of type str but was {type(class_name)}")
        if not isinstance(properties, (list, str)):
            raise TypeError(f"properties must be of type str or \
                            list of str but was {type(properties)}")
        if isinstance(properties, str):
            properties = [properties]

        self._class_name = class_name
        self._properties = properties
        self._where: Optional[WhereFilter] = None  # To store the where filter if it is added
        self._limit: Optional[str] = None  # To store the limit filter if it is added
        self._near_text: Optional[NearText] = None # To store the nearText clause if it is added
        self._near_vector: Optional[NearVector] = None # To store the nearText clause if it is added
        self._contains_filter = False  # true if any filter is added

    def with_where(self, content: dict) -> 'GetBuilder':
        """
        Set `where` filter.

        Parameters
        ----------
        content : dict
            The content of the where filter to set.

        Returns
        -------
        weaviate.gql.get.GetBuilder
            Updated GetBuilder.
        """

        self._where = WhereFilter(content)
        self._contains_filter = True
        return self

    def with_near_text(self, content: dict) -> 'GetBuilder':
        """
        Set `nearText` filter.

        Parameters
        ----------
        content : dict
            The content of the nearText filter to set.

        Returns
        -------
        weaviate.gql.get.GetBuilder
            Updated GetBuilder.

        Raises
        ------
        AttributeError
            If 'nearVector' was already set.
        """

        if self._near_vector is not None:
            raise AttributeError("Cannot use both 'nearText' and 'nearVector' filters!")
        self._near_text = NearText(content)
        self._contains_filter = True
        return self

    def with_near_vector(self, vector: list) -> 'GetBuilder':
        """
        Set `nearVector` filter.

        Parameters
        ----------
        vector : list
            The vector of the nearVector filter to set.

        Returns
        -------
        weaviate.gql.get.GetBuilder
            Updated GetBuilder.

        Raises
        ------
        AttributeError
            If 'nearText' was already set.
        """

        if self._near_text is not None:
            raise AttributeError("Cannot use both 'nearText' and 'nearVector' filters!")
        self._near_vector = NearVector(vector)
        self._contains_filter = True
        return self

    def with_limit(self, limit: int) -> 'GetBuilder':
        """
        The limit of objects returned.

        Parameters
        ----------
        limit : dict
            The max number of objects returned.

        Returns
        -------
        weaviate.gql.get.GetBuilder
            Updated GetBuilder.

        Raises
        ------
        ValueError
            If 'limit' is non-positive.
        """

        if limit < 1:
            raise ValueError('limit cannot be negative (limit >=1).')

        self._limit = f'limit: {limit}'
        self._contains_filter = True
        return self

    def build(self) -> str:
        """
        Build query filter as a string.

        Returns
        -------
        str
            The GraphQL query as a string.
        """

        query = f'{{Get{{{self._class_name}'
        if self._contains_filter:
            query += '('
            if self._where is not None:
                query = query + str(self._where)
            if self._limit is not None:
                query = query + self._limit
            if self._near_vector is not None:
                query = query + str(self._near_vector)
            if self._near_text is not None:
                query = query + str(self._near_text)
            query += ')'
        query = query + f'{{{" ".join(self._properties)}}}}}}}'

        return query

    def do(self) -> dict:
        """
        Builds and runs the query.

        Returns
        -------
        dict
            The response of the query.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        query = self.build()

        try:
            response = self._connection.run_rest("/graphql", REST_METHOD_POST, {"query": query})
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, query was not successful.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])

        if response.status_code == 200:
            return response.json()  # success
        raise UnexpectedStatusCodeException("Query was not successful", response)



import time
import urllib
from typing import Tuple, Optional
import requests
from weaviate.exceptions import AuthenticationFailedException
from weaviate.connect.constants import *
from weaviate.auth import AuthCredentials
from .util import get_epoch_time

WEAVIATE_REST_API_VERSION_PATH = "/v1"


class Connection:
    """
    Connection class used to communicate to a weaviate instance.
    """
    def __init__(self,
            url: str,
            auth_client_secret: Optional[AuthCredentials]=None,
            timeout_config: Optional[Tuple[int, int]]=None
        ):
        """
        Initialize a Connection class instance.

        Parameters
        ----------
        url : str
            URL to a running weaviate instance.
        auth_client_secret : weaviate.auth.AuthCredentials, optional
            User login credentials to a weaviate instance, by default None
        timeout_config : tuple(int, int), optional
            Set the timeout config as a tuple of (retries, time out seconds),
            by default None.
        """

        self.url = url + WEAVIATE_REST_API_VERSION_PATH  # e.g. http://localhost:80/v1
        if timeout_config is None:
            self.timeout_config = (2, 20)
        else:
            self.timeout_config = timeout_config

        self.auth_expires = 0  # unix time when auth expires
        self.auth_bearer = 0
        self.auth_client_secret = auth_client_secret

        self.is_authentication_required = False
        try:
            request = requests.get(
                self.url + "/.well-known/openid-configuration",
                headers={"content-type": "application/json"},
                timeout=(30, 45)
                )
        except Exception:
            pass
        else:
            if request.status_code == 200:
                if isinstance(auth_client_secret, AuthCredentials):
                    self.is_authentication_required = True
                    self._refresh_authentication()
                else:
                    raise ValueError(f"No login credentials provided. The weaviate instance at \
                        {url} requires login credential, use argument 'auth_client_secret'.")

    # Requests a new bearer
    def _refresh_authentication(self) -> None:
        """
        Request a new bearer.

        Raises
        ------
        AuthenticationFailedException
            If cannot connect to weaviate.
        AuthenticationFailedException
            If cannot authenticate http status not ok.
        AuthenticationFailedException
            If cannot connect to the third party authentication service.
        AuthenticationFailedException
            If status not OK in connection to the third party authentication service.
        AuthenticationFailedException
            If the grant_types supported by the thirdparty authentication service are insufficient.
        AuthenticationFailedException
            If unable to get a OAuth token from server.
        AuthenticationFailedException
            If authtentication access denied.
        """

        if (self.auth_expires - 2) < get_epoch_time():  # -2 for some lagtime
            # collect data for the request
            try:
                request = requests.get(
                    self.url + "/.well-known/openid-configuration",
                    headers={"content-type": "application/json"},
                    timeout=(30, 45)
                    )
            except urllib.error.HTTPError as error:
                raise AuthenticationFailedException("Cant connect to weaviate") from error
            if request.status_code != 200:
                raise AuthenticationFailedException("Cant authenticate http status not ok")

            # Set the client ID
            client_id = request.json()['clientId']

            # request additional information
            try:
                request_third_part = requests.get(
                    request.json()['href'],
                    headers={"content-type": "application/json"},
                    timeout=(30, 45)
                    )
            except urllib.error.HTTPError as error:
                raise AuthenticationFailedException(
                    "Can't connect to the third party authentication service. \
                        Validate that it's running") from error
            if request_third_part.status_code != 200:
                raise AuthenticationFailedException(
                    "Status not OK in connection to the third party authentication service")

            # Validate third part auth info
            if 'client_credentials' not in request_third_part.json()['grant_types_supported']:
                raise AuthenticationFailedException(
                    "The grant_types supported by the thirdparty authentication service are \
                        insufficient. Please add 'client_credentials'")

            request_body = self.auth_client_secret.get_credentials()
            request_body["client_id"] = client_id

            # try the request
            try:
                request = requests.post(
                    request_third_part.json()['token_endpoint'],
                    request_body,
                    timeout=(30, 45)
                    )
            except urllib.error.HTTPError as error:
                raise AuthenticationFailedException(
                    "Unable to get a OAuth token from server. Are the credentials \
                        and URLs correct?") from error

            # sleep to process
            time.sleep(0.125)

            if request.status_code == 401:
                raise AuthenticationFailedException(
                    "Authtentication access denied. Are the credentials correct?"
                )
            self.auth_bearer = request.json()['access_token']
            self.auth_expires = int(get_epoch_time() + request.json()['expires_in'] - 2)


    def _get_request_header(self) -> dict:
        """
        Returns the correct headers for a request.

        Returns
        -------
        dict
            Request header as a dict.
        """

        header = {"content-type": "application/json"}

        if self.is_authentication_required:
            self._refresh_authentication()
            header["Authorization"] = "Bearer " + self.auth_bearer

        return header

    def run_rest(self,
            path: str,
            rest_method: int,
            weaviate_object: dict=None,
            params: dict=None
        ) -> requests.Response:
        """
        Make a request to the weaviate instance.

        Parameters
        ----------
        path : str
            Sub-path to the weaviate resources. Must be a valid weaviate sub-path.
            e.g. /meta or /objects, without version.
        rest_method : int
            Type of the rest API request. Is defined through a constant given in
            the package e.g. REST_METHOD_GET.
        weaviate_object : dict, optional
            Object is used as payload, by default None
        params : dict, optional
            Additional request prameters, by default None

        Returns
        -------
        requests.Response
            The response if request was successful.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the request could not be made.
            (from requests.'method' calls)
        """

        if params is None:
            params = {}
        request_url = self.url+path

        if rest_method == REST_METHOD_GET:
            response = requests.get(
                url=request_url,
                headers=self._get_request_header(),
                timeout=self.timeout_config,
                params=params
                )
        elif rest_method == REST_METHOD_PUT:
            response = requests.put(
                url=request_url,
                json=weaviate_object,
                headers=self._get_request_header(),
                timeout=self.timeout_config
                )
        elif rest_method == REST_METHOD_POST:
            response = requests.post(
                url=request_url,
                json=weaviate_object,
                headers=self._get_request_header(),
                timeout=self.timeout_config
                )
        elif rest_method == REST_METHOD_PATCH:
            response = requests.patch(
                url=request_url,
                json=weaviate_object,
                headers=self._get_request_header(),
                timeout=self.timeout_config
                )
        elif rest_method == REST_METHOD_DELETE:
            response = requests.delete(
                url=request_url,
                json=weaviate_object,
                headers=self._get_request_header(),
                timeout=self.timeout_config
                )
        else:
            print("Not yet implemented rest method called")
            response = None
        return response



import sys
import time
from typing import Dict, Any
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from weaviate.connect import REST_METHOD_POST, Connection


class ConfigBuilder:
    """
    ConfigBuild class that is used to configure a classification process.
    """

    def __init__(self, connection: Connection, classification: 'Classification'):
        """
        Initialize a ConfiBuilder class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        classification : weaviate.classification.Classification
            Classification object to be configured using this ConfigBuilder
            instance.
        """

        self._connection = connection
        self._classification = classification
        self._config: Dict[str, Any] = {}
        self._wait_for_completion = False

    def with_type(self, type: str) -> 'ConfigBuilder':
        """
        Set classification type.

        Parameters
        ----------
        type : str
            Type of the desired classification.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        self._config["type"] = type
        return self

    def with_k(self, k: int) -> 'ConfigBuilder':
        """
        Set k number for the kNN.

        Parameters
        ----------
        k : int
            Number of objects to use to make a classification guess.
            (For kNN)

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        self._config["settings"] = {
            "k" : k
        }
        return self

    def with_class_name(self, class_name: str) -> 'ConfigBuilder':
        """
        What Object type to classify.

        Parameters
        ----------
        class_name : str
            Name of the class to be classified.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        self._config["class"] = class_name
        return self

    def with_classify_properties(self, classify_properties: list) -> 'ConfigBuilder':
        """
        Set the classify properties.

        Parameters
        ----------
        classify_properties: list
            A list of properties to clasify.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        self._config["classifyProperties"] = classify_properties
        return self

    def with_based_on_properties(self, based_on_properties: list) -> 'ConfigBuilder':
        """
        Set properties to build the classification on.

        Parameters
        ----------
        based_on_properties: list
            A list of properties to classify on.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        self._config["basedOnProperties"] = based_on_properties
        return self

    def with_source_where_filter(self, filter: dict) -> 'ConfigBuilder':
        """
        Set Source 'where' Filter.

        Parameters
        ----------
        filter : dict
            Filter to use, as a dict.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        if "filters" not in self._config:
            self._config["filters"] = {}
        self._config["filters"]["sourceWhere"] = filter
        return self

    def with_training_set_where_filter(self, filter: dict) -> 'ConfigBuilder':
        """
        Set Training set 'where' Filter.

        Parameters
        ----------
        filter : dict
            Filter to use, as a dict.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        if "filters" not in self._config:
            self._config["filters"] = {}
        self._config["filters"]["trainingSetWhere"] = filter
        return self

    def with_target_where_filter(self, filter: dict) -> 'ConfigBuilder':
        """
        Set Target 'where' Filter.

        Parameters
        ----------
        filter : dict
            Filter to use, as a dict.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        if "filters" not in self._config:
            self._config["filters"] = {}
        self._config["filters"]["targetWhere"] = filter
        return self

    def with_wait_for_completion(self) -> 'ConfigBuilder':
        """
        Wait for completion.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        self._wait_for_completion = True
        return self

    def with_settings(self, settings: dict) -> 'ConfigBuilder':
        """
        Set settings for the classification. NOTE if you are using 'kNN'
        the value 'k' can be set by this method or by 'with_k'.
        This method overwrites any previously set settings, like 'k' from
        the 'with_k' method.

        Parameters
        ----------
        settings: dict
            Additional settings to be set/overwritten.

        Returns
        -------
        ConfigBuilder
            Updated ConfigBuilder.
        """

        self._config["settings"] = settings
        return self

    def _validate_config(self) -> None:
        """
        Validate the current classification configuration.

        Raises
        ------
        ValueError
            If a mandatory field is not set.
        """

        required_fields = ["type", "class", "basedOnProperties", "classifyProperties"]
        for field in required_fields:
            if field not in self._config:
                raise ValueError(f"{field} is not set for this classification")

        if "settings" in self._config:
            if not isinstance(self._config["settings"], dict):
                TypeError('"settings" should be of type dict')

        if self._config["type"] == "knn":
            if "k" not in self._config.get("settings", []):
                raise ValueError("k is not set for this classification")

    def _start(self) -> dict:
        """
        Start the classification based on the configuration set.

        Returns
        -------
        dict
            Classification result.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            Unexpected error.
        """

        try:
            response = self._connection.run_rest("/classifications", REST_METHOD_POST, self._config)
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                        + ' Connection error, classification may not started.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 201:
            return response.json()
        raise UnexpectedStatusCodeException("Start classification", response)

    def do(self) -> dict:
        """
        Start the classification.

        Returns
        -------
        dict
            Classification result.
        """

        self._validate_config()

        response = self._start()
        if not self._wait_for_completion:
            return response

        # wait for completion
        classification_uuid = response["id"]
        #print(classification_uuid)
        while self._classification.is_running(classification_uuid):
            time.sleep(2.0)
        return self._classification.get(classification_uuid)



import sys
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from weaviate.schema.validate_schema import check_property
from weaviate.util import _get_dict_from_object
from weaviate.connect import REST_METHOD_POST, REST_METHOD_DELETE, Connection


class Property:
    """
    Property class used to create object properties.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Property class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    def create(self, schema_class_name: str, schema_property: dict) -> None:
        """
        Create a class property.

        Parameters
        ----------
        schema_class_name : str
            The name of the class in the schema to which the property
            should be added.
        schema_property : dict
            The property that should be added.

        Raises
        ------
        TypeError
            If 'schema_class_name' is of wrong type.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.SchemaValidationException
            If the 'schema_property' is not valid.
        """

        if not isinstance(schema_class_name, str):
            raise TypeError("Class name must be of type str but is", type(schema_class_name))

        loaded_schema_property = _get_dict_from_object(schema_property)

        # check if valid property
        check_property(loaded_schema_property)

        path = f"/schema/{schema_class_name}/properties"
        try:
            response = self._connection.run_rest(path, REST_METHOD_POST, loaded_schema_property)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, property may not have \
                                        been created properly.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Add property to class", response)

    def _delete(self, schema_class_name: str, schema_property_name: str) -> None:
        """
        This function is currently deactivated because it is not available in weaviate yet:
        https://github.com/semi-technologies/weaviate/issues/973

        Parameters
        ----------
        schema_class_name : str
            The class the property that is being removed is a part of.
        schema_property_name : str
            The name of the property that should be removed.

        Raises
        ------
        TypeError
            If argument/s is/are of wrong type/s.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.SchemaValidationException
            If the 'schema_property' is not valid.
        """

        if not isinstance(schema_class_name, str):
            raise TypeError("Class name should be str not", type(schema_class_name))
        if not isinstance(schema_property_name, str):
            raise TypeError("Property name should be str not", type(schema_property_name))

        path = f"/schema/{schema_class_name}/properties/{schema_property_name}"
        try:
            response = self._connection.run_rest(path, REST_METHOD_DELETE)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, property may not have \
                                        been deleted properly.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Delete property from class", response)



from setuptools import setup
from builtins import open

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'readme.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name="weaviate-client",
      version="2.0.0rc1",
      description="A python native weaviate client",
      long_description=long_description,
      long_description_content_type='text/markdown',
      author="SeMI Technologies",
      author_email="hello@semi.technology",
      packages=["weaviate", "weaviate.connect", "weaviate.tools", "weaviate.rdf", "weaviate.schema", "weaviate.schema.properties", "weaviate.batch", "weaviate.classification", "weaviate.contextionary", "weaviate.data", "weaviate.data.references", "weaviate.gql"],
      python_requires='>=3.6',
      install_requires=[
        "requests>=2.22.0",
        "validators>=0.14.0",
        "rdflib==5.0.0"]),



from typing import Optional
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from .connect import Connection, REST_METHOD_GET
from .classification import Classification
from .schema import Schema
from .contextionary import Contextionary
from .batch import Batch
from .data import DataObject
from .gql import Query
from .client_config import ClientConfig
from .auth import AuthCredentials


class Client:
    """
    A python native weaviate client.
    """

    def __init__(self,
            url: str,
            auth_client_secret: AuthCredentials=None,
            client_config: ClientConfig=None
        ):
        """
        Initialize a Client class instance.

        Parameters
        ----------
        url : str
            The URL to the weaviate instance.
        auth_client_secret : weaviate.AuthCredentials, optional
            Authentification client secret, by default None.
        client_config : weaviate.ClientConfig, optional
            Gives additional optimization parameters for the client.
            Uses default parameters if omitted, by default None.

        Raises
        ------
        TypeError
            If arguments are of a wrong data type.
        ValueError
            If 'client_config.timeout_config' is not a tuple of 2.
        """

        if not isinstance(url, str):
            raise TypeError("URL is expected to be string but is "+str(type(url)))
        if url.endswith("/"):
            # remove trailing slash
            url = url[:-1]

        if client_config is not None:
            # Check the input
            if (not isinstance(client_config.timeout_config, tuple)) or\
                    (not isinstance(client_config.timeout_config[0], int)) or\
                    (not isinstance(client_config.timeout_config[1], int)):
                raise TypeError("ClientConfig.timeout_config must be tupel of int")
            if len(client_config.timeout_config) != 2:
                raise ValueError("ClientConfig.timeout_config must be of length 2")
        else:
            # Create the default config
            client_config = ClientConfig()

        self._connection = Connection(url=url,
                                    auth_client_secret=auth_client_secret,
                                    timeout_config=client_config.timeout_config)

        self.classification = Classification(self._connection)
        self.schema = Schema(self._connection)
        self.contextionary = Contextionary(self._connection)
        self.batch = Batch(self._connection)
        self.data_object = DataObject(self._connection)
        self.query = Query(self._connection)

    def is_ready(self) -> bool:
        """
        Ping weaviates ready state

        Returns
        -------
        bool
            True if weaviate is ready to accept requests,
            False otherwise.
        """

        try:
            response = self._connection.run_rest("/.well-known/ready", REST_METHOD_GET)
            if response.status_code == 200:
                return True
            return False
        except RequestsConnectionError:
            return False

    def is_live(self) -> bool:
        """
        Ping weaviates live state.

        Returns
        --------
        bool
            True if weaviate is live and should not be killed,
            False otherwise.
        """

        response = self._connection.run_rest("/.well-known/live", REST_METHOD_GET)
        if response.status_code == 200:
            return True
        return False

    def get_meta(self) -> dict:
        """
        Get the meta endpoint description of weaviate.

        Returns
        -------
        dict
            The dict describing the weaviate configuration.

        Raises
        ------
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        response = self._connection.run_rest("/meta", REST_METHOD_GET)
        if response.status_code == 200:
            return response.json()
        raise UnexpectedStatusCodeException("Meta endpoint", response)

    def get_open_id_configuration(self) -> Optional[dict]:
        """
        Get the openid-configuration.

        Returns
        -------
        dict
            The configuration or None if not configured.

        Raises
        ------
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        response = self._connection.run_rest("/.well-known/openid-configuration", REST_METHOD_GET)
        if response.status_code == 200:
            return response.json()
        if response.status_code == 404:
            return None
        raise UnexpectedStatusCodeException("Meta endpoint", response)



import sys
import validators
from weaviate.exceptions import UnexpectedStatusCodeException, RequestsConnectionError
from weaviate.connect import REST_METHOD_GET, Connection
from .config_builder import ConfigBuilder

class Classification:
    """
    Classification class used to schedule and/or check the status of
    a classification process of weaviate objects.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Classification class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    def schedule(self) -> ConfigBuilder:
        """
        Schedule a Classification of the Objects within weaviate.

        Returns
        -------
        weaviate.classification.config_builder.ConfigBuilder
            A ConfigBuilder that should be configured to the desired
            classification task
        """

        return ConfigBuilder(self._connection, self)

    def get(self, classification_uuid: str) -> dict:
        """
        Polls the current state of the given classification.

        Parameters
        ----------
        classification_uuid : str
            Identifier of the classification.

        Returns
        -------
        dict
            A dict containing the weaviate answer.

        Raises
        ------
        ValueError
            If not a proper uuid.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if not validators.uuid(classification_uuid):
            raise ValueError("Given UUID does not have a proper form")

        try:
            response = self._connection.run_rest("/classifications/" + classification_uuid,\
                                                                         REST_METHOD_GET)
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                    + ' Connection error, classification status could not be retrieved.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 200:
            return response.json()
        raise UnexpectedStatusCodeException("Get classification status", response)

    def is_complete(self, classification_uuid: str) -> bool:
        """
        Checks if a started classification job has completed.

        Parameters
        ----------
        classification_uuid : str
            Identifier of the classification.

        Returns
        -------
        bool
            True if given classification has finished, False otherwise.
        """

        return self._check_status(classification_uuid, "completed")

    def is_failed(self, classification_uuid: str) -> bool:
        """
        Checks if a started classification job has failed.

        Parameters
        ----------
        classification_uuid : str
            Identifier of the classification.

        Returns
        -------
        bool
            True if the classification failed, False otherwise.
        """

        return self._check_status(classification_uuid, "failed")

    def is_running(self, classification_uuid: str) -> bool:
        """
        Checks if a started classification job is running.

        Parameters
        ----------
        classification_uuid : str
            Identifier of the classification.

        Returns
        -------
        bool
            True if the classification is running, False otherwise.
        """

        return self._check_status(classification_uuid, "running")

    def _check_status(self,
            classification_uuid: str,
            status: str
        ) -> bool:
        """
        Check for a status of a classification.

        Parameters
        ----------
        classification_uuid : str
            Identifier of the classification.
        status : str
            Status to check for.

        Returns
        -------
        bool
            True if 'status' is satisfied, False otherwise.
        """

        try:
            response = self.get(classification_uuid)
        except RequestsConnectionError:
            return False
        if response["status"] == status:
            return True
        return False



import sys
from typing import Union, Optional, List
import validators
from requests import Response
from weaviate.connect import REST_METHOD_POST
from weaviate.connect import REST_METHOD_PATCH
from weaviate.connect import REST_METHOD_PUT
from weaviate.connect import REST_METHOD_GET
from weaviate.connect import REST_METHOD_DELETE
from weaviate.connect import Connection
from weaviate.exceptions import ObjectAlreadyExistsException
from weaviate.exceptions import RequestsConnectionError
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.util import _get_dict_from_object
from weaviate.data.references import Reference


class DataObject:
    """
    DataObject class used to manipulate object to/from weaviate.
    """
    def __init__(self, connection: Connection):
        """
        Initialize a DataObject class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection
        self.reference = Reference(self._connection)

    def create(self,
            data_object: Union[dict, str],
            class_name: str,
            uuid: str=None,
            vector_weights: dict=None
        ) -> str:
        """
        Takes a dict describing the object and adds it to weaviate.

        Parameters
        ----------
        data_object : dict or str
            Object to be added.
            If type is str it should be either an URL or a file.
        class_name : str
            Class name associated with the object given.
        uuid : str, optional
            Object will be created under this uuid if it is provided.
            Otherwise weaviate will generate a uuid for this object,
            by default None.
        vector_weights : dict, optional
            Influence the weight of words on object creation.
            Default is None for no influence.

        Returns
        -------
        str
            Returns the UUID of the created object if successful.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        weaviate.ThingAlreadyExistsException
            If an object with the given uuid already exists within weaviate.
        weaviate.UnexpectedStatusCodeException
            If creating the object in weavate failed with a different reason,
            more information is given in the exception.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        """

        loaded_data_object = _get_dict_from_object(data_object)
        if not isinstance(class_name, str):
            raise TypeError("Expected class_name of type str but was: "\
                            + str(type(class_name)))

        weaviate_obj = {
            "class": class_name,
            "properties": loaded_data_object
        }
        if uuid is not None:
            if not isinstance(uuid, str):
                raise TypeError("Expected uuid to be of type str but was: "\
                                + str(type(uuid)))
            if not validators.uuid(uuid):
                raise ValueError("Given uuid does not have a valid form")

            weaviate_obj["id"] = uuid

        if vector_weights is not None:
            if not isinstance(vector_weights, dict):
                raise TypeError("Expected vector_weights to be of type dict but was "\
                                + str(type(vector_weights)))

            weaviate_obj["vectorWeights"] = vector_weights

        path = "/objects"
        try:
            response = self._connection.run_rest(path, REST_METHOD_POST, weaviate_obj)
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                    + ' Connection error, object was not added to weaviate.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])

        if response.status_code == 200:
            return str(response.json()["id"])

        object_does_already_exist = False
        try:
            if 'already exists' in response.json()['error'][0]['message']:
                object_does_already_exist = True
        except KeyError:
            pass
        except Exception as error:
            message = str(error)\
                    + ' Unexpected exception please report this excetpion in an issue.'
            raise type(error)(message).with_traceback(sys.exc_info()[2])

        if object_does_already_exist:
            raise ObjectAlreadyExistsException(str(uuid))
        raise UnexpectedStatusCodeException("Creating object", response)

    def merge(self,
            data_object: Union[dict, str],
            class_name: str,
            uuid: str
        ) -> None:
        """
        Merge the given object with the already existing object in weaviate.
        Overwrites all given fields.

        Parameters
        ----------
        data_object : dict or str
            The object states the fields that should be updated.
            Fields not stated by object will not be changed.
            Fields that are None will not be changed.
            If type is str it should be either an URL or a file.
        class_name : str
            The class name of the object.
        uuid : str
            The ID of the object that should be changed.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none successful status.
        """

        object_dict = _get_dict_from_object(data_object)

        if not isinstance(class_name, str):
            raise TypeError("Class must be type str")
        if not isinstance(uuid, str):
            raise TypeError("UUID must be type str")
        if not validators.uuid(uuid):
            raise ValueError("Not a proper UUID")

        payload = {
            "id": uuid,
            "class": class_name,
            "properties": object_dict
        }

        path = f"/objects/{uuid}"

        try:
            response = self._connection.run_rest(path, REST_METHOD_PATCH, payload)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, object was not patched.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 204:
            # Successful merge
            return
        raise UnexpectedStatusCodeException("PATCH merge of object not successful", response)

    def update(self,
            data_object: Union[dict, str],
            class_name: str,
            uuid: str
        ) -> None:
        """
        Update an already existing object with the given data object.
        Does not keep unset values.

        Parameters
        ----------
        data_object : dict or str
            Describes the new values. It may be an URL or path to a json
            or a python dict describing the new values.
        class_name : str
            Name of the class of the object that should be updated.
        uuid : str
            The UUID of the object that should be changed.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        parsed_object = _get_dict_from_object(data_object)

        weaviate_obj = {
            "id": uuid,
            "class": class_name,
            "properties": parsed_object
        }

        path = f"/objects/{uuid}"
        try:
            response = self._connection.run_rest(path, REST_METHOD_PUT, weaviate_obj)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error, object was not updated.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 200:
            # Successful update
            return
        raise UnexpectedStatusCodeException("Update object", response)

    def get_by_id(self,
            uuid: str,
            additional_properties: List[str]=None,
            with_vector: bool=False
        ) -> Optional[dict]:
        """
        Get an object as dict.

        Parameters
        ----------
        uuid : str
            The identifier of the object that should be retrieved.
        additional_properties : list of str, optional
            List of additional properties that should be included in the request,
            by default None
        with_vector: bool
            If True the `vector` property will be returned too,
            by default False.

        Returns
        -------
        dict or None
            dict in case the object exists.
            None in case the object does not exist.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        response = self._get_object_response(uuid, additional_properties, with_vector)

        if response.status_code == 200:
            return response.json()
        if response.status_code == 404:
            return None
        raise UnexpectedStatusCodeException("Get object", response)

    def get(self,
            additional_properties: List[str]=None,
            with_vector: bool=False
        ) -> List[dict]:
        """
        Gets all objects.

        Parameters
        ----------
        additional_properties : list of str, optional
            list of additional properties that should be included in the request,
            by default None
        with_vector: bool
            If True the `vector` property will be returned too,
            by default False.

        Returns
        -------
        list of dicts
            A list of all objects. If no objects where found the list is empty.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        params = _get_params(additional_properties, with_vector)

        path = "/objects"

        try:
            response = self._connection.run_rest(path, REST_METHOD_GET, params=params)
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error when getting objects'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])

        if response.status_code == 200:
            return response.json()
        raise UnexpectedStatusCodeException("Get object", response)

    def _get_object_response(self,
            object_uuid: str,
            additional_properties: List[str],
            with_vector: bool
        ) -> Response:
        """
        Retrieve an object from weaviate.

        Parameters
        ----------
        object_uuid : str
            The identifier of the object that should be retrieved.
        additional_properties : list of str, optional
            Defines the additional properties that should be included in the result.
        with_vector: bool
            If True the `vector` property will be returned too.

        Returns
        -------
        requests.Response
            Respose object.

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        """

        params = _get_params(additional_properties, with_vector)

        if not isinstance(object_uuid, str):
            object_uuid = str(object_uuid)
        try:
            response = self._connection.run_rest(
                "/objects/" + object_uuid,
                REST_METHOD_GET,
                params=params
                )
        except RequestsConnectionError as conn_err:
            message = str(conn_err) + ' Connection error not sure if object exists'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        return response

    def delete(self, uuid: str) -> None:
        """
        Delete an existing object from weaviate.

        Parameters
        ----------
        uuid : str
            The ID of the object that should be deleted.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If parameter has the wrong type.
        ValueError
            If uuid is not properly formed.
        """

        if not isinstance(uuid, str):
            raise TypeError("UUID must be type str")
        if not validators.uuid(uuid):
            raise ValueError("UUID does not have proper form")

        try:
            response = self._connection.run_rest("/objects/" + uuid, REST_METHOD_DELETE)
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                    + ' Connection error, object could not be deleted.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])
        if response.status_code == 204:
            # Successfully deleted
            return
        raise UnexpectedStatusCodeException("Delete object", response)

    def exists(self, uuid: str) -> bool:
        """
        Check if the object exist in weaviate.

        Parameters
        ----------
        uuid : str
            The UUID of the object that may or may not exist within weaviate.

        Returns
        -------
        bool
            True if object exists, False otherwise.

        Raises
        ------
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If parameter has the wrong type.
        ValueError
            If uuid is not properly formed.
        """

        response = self._get_object_response(uuid, None, False)

        if response.status_code == 200:
            return True
        if response.status_code == 404:
            return False
        raise UnexpectedStatusCodeException("Object exists", response)

    def validate(self,
            data_object: Union[dict, str],
            class_name: str,
            uuid: str
        ) -> dict:
        """
        Validate an object against weaviate.

        Parameters
        ----------
        data_object : dict or str
            Object to be validated.
            If type is str it should be either an URL or a file.
        class_name : str
            Name of the class of the object that should be validated.
        uuid : str
            The UUID of the object that shoudl be validated against weaviate.

        Returns
        -------
        dict
            Validation result. E.g.
            {
                "valid": bool,
                "error": None or list
            }

        Raises
        ------
        TypeError
            If argument is of wrong type.
        ValueError
            If argument contains an invalid value.
        weaviate.UnexpectedStatusCodeException
            If validating the object against weavate failed with a different reason.
        requests.exceptions.ConnectionError
            If the network connection to weaviate fails.
        """

        if not isinstance(uuid, str):
            raise TypeError("UUID must be of type str")

        loaded_data_object = _get_dict_from_object(data_object)
        if not isinstance(class_name, str):
            raise TypeError(f"Expected class_name of type str but was: {type(class_name)}")

        weaviate_obj = {
            "id": uuid,
            "class": class_name,
            "properties": loaded_data_object
        }

        path = "/objects/validate"
        try:
            response = self._connection.run_rest(path, REST_METHOD_POST, weaviate_obj)
        except RequestsConnectionError as conn_err:
            message = str(conn_err)\
                    + ' Connection error, object was not validated against weaviate.'
            raise type(conn_err)(message).with_traceback(sys.exc_info()[2])

        result: dict = {
            "error": None
        }

        if response.status_code == 200:
            result["valid"] = True
            return result
        if response.status_code == 422:
            result["valid"] = False
            result["error"] = response.json()["error"]
            return result
        raise UnexpectedStatusCodeException("Validate object", response)


def _get_params(additional_properties: Optional[List[str]], with_vector: bool) -> dict:
    """
    Get underscor properties in the format accepted by weaviate.

    Parameters
    ----------
    additional_properties : list of str or None
        A list of additional properties or None.
    with_vector: bool
        If True the `vector` property will be returned too.

    Returns
    -------
    dict
        A dictionary including weaviate-accepted additional properties
        and/or `vector` property.

    Raises
    ------
    TypeError
        If 'additional_properties' is not of type list.
    """

    params = {}
    if additional_properties is not None:
        if not isinstance(additional_properties, list):
            raise TypeError(f"Additional properties must be of type list \
                                but are {type(additional_properties)}")
        params['include'] = ",".join(additional_properties)

    if with_vector:
        if 'include' in params:
            params['include'] = params['include'] + ',vector'
        else:
            params['include'] = 'vector'
    return params
